@@ -1,15 +1,18 @@
 # -*- coding: utf-8 -*-
 import os
 import platform
+import re
 import subprocess
-from typing import List, Tuple
+from typing import List, Tuple, Union
 
 import click
 from kubernetes import client, config, watch
 from kubernetes.client.rest import ApiException
+from kubernetes.stream import stream
 from urllib3.exceptions import MaxRetryError
 
 import src.cli.console as console
+from src import settings
 from src.local.exceptions import UnikubeClusterUnavailableError
 
 
@@ -128,10 +131,15 @@ class KubeCtl(CMDWrapper):
         output = process.stdout.read()
         return output
 
-    def exec_pod(self, pod, namespace, command, interactive=False):
+    def exec_pod(self, pod, namespace, command, interactive=False, container=None):
         arguments = ["exec", "-it", pod]
+
         if namespace:
             arguments += ["--namespace", namespace]
+
+        if container:
+            arguments += ["--container", container]
+
         arguments += ["--", command]
         if interactive:
             console.info(f"Running '{command}' (interactively) on: {pod}. Please press Ctrl+D to exit.")
@@ -165,11 +173,22 @@ class Docker(CMDWrapper):
             )
 
     def check_running(self, name):
+        """Checks whether an image or a specific container is running."""
         arguments = ["ps"]
         process = self._execute(arguments)
         output = process.stdout.read()
         return name in output
 
+    def daemon_active(self):
+        """Checks whether docker daemon is running.
+
+        Based on docker documentation (https://docs.docker.com/config/daemon/#check-whether-docker-is-running).
+        `docker info` exists with non-zero exit code when docker is not running.
+        """
+        arguments = ["info"]
+        process = self._execute(arguments)
+        return process.returncode == 0
+
     def exec(self, container, command, interactive=False):
         arguments = ["exec", "-it", container, command]
         if interactive:
@@ -195,6 +214,15 @@ class Docker(CMDWrapper):
         output = process.stdout.read()
         return output.strip()
 
+    def image_exists(self, name):
+        arguments = ["images", "-q", name]
+        process = self._execute(arguments)
+        output = process.stdout.read()
+        if output.strip():
+            return True
+        else:
+            return False
+
 
 class Telepresence(KubeCtl):
     base_command = "telepresence"
@@ -258,10 +286,9 @@ class Telepresence(KubeCtl):
 
     def uninstall(self, deployment, namespace=None, silent=False):
         arguments = ["uninstall", "--agent", deployment]
+        arguments.append(deployment)
         if namespace:
-            arguments.append(f"{deployment}-{namespace}")
-        else:
-            arguments.append(deployment)
+            arguments += ["-n", namespace]
         console.debug(arguments)
         process = self._execute(arguments)
         if not silent and process.returncode and process.returncode != 0:
@@ -316,6 +343,22 @@ class Telepresence(KubeCtl):
         swapped = any(filter(lambda x: x[0] == deployment and x[1] == "intercepted", deployments))
         return swapped
 
+    def intercept_count(self) -> Union[int, None]:
+        arguments = ["status"]
+        process = self._execute(arguments)
+        status = process.stdout.readlines()
+
+        # parse intercept count
+        try:
+            intercept_line = status[15]
+            match = re.findall("[ ]{1,}Intercepts[ ]{1,}:(.*)[ ]{1,}total", intercept_line)
+            intercept_count = int(match[0])
+        except Exception as e:
+            console.debug(e)
+            intercept_count = None
+
+        return intercept_count
+
 
 class KubeAPI(object):
     def __init__(self, provider_data, deck=None):
@@ -337,13 +380,20 @@ class KubeAPI(object):
         except ApiException:
             return False
 
+    def delete_pod(self, pod_name):
+        delete_options = client.V1DeleteOptions()
+        try:
+            return self._core_api.delete_namespaced_pod(name=pod_name, namespace=self._namespace, body=delete_options)
+        except MaxRetryError:
+            raise UnikubeClusterUnavailableError
+
     def get_pods(self):
         try:
             return self._core_api.list_namespaced_pod(self._namespace, watch=False)
         except MaxRetryError:
             raise UnikubeClusterUnavailableError
 
-    def get_pod(self, pod_name):
+    def get_pod(self, pod_name: str):
         pods = self.get_pods()
         for pod in pods.items:
             if pod.metadata.name == pod_name:
@@ -377,3 +427,49 @@ class KubeAPI(object):
         if name:
             ret = next(filter(lambda x: x.metadata.name == name, ret.items))
         return ret
+
+    def get_serviceaccount_tokens(self, app_name):
+        cmd = [
+            f"cat {settings.SERVICE_TOKEN_FILENAME}",
+            f"cat {settings.SERVICE_CERT_FILENAME}",
+        ]
+
+        pods = self.get_pods()
+        target = None
+        for pod in pods.items:
+            # match the first pod from this app
+            if pod.metadata.name.startswith(app_name):
+                target = pod
+                break
+        if target is None:
+            print(app_name)
+            return None
+        # select the first container
+        container_name = target.spec.containers[0].name
+        response = stream(
+            self._core_api.connect_get_namespaced_pod_exec,
+            target.metadata.name,
+            self._namespace,
+            container=container_name,
+            command="/bin/sh",
+            stderr=True,
+            stdin=True,
+            stdout=True,
+            tty=False,
+            _preload_content=False,
+        )
+        std_out = []
+        std_err = []
+        while response.is_open():
+            response.update(timeout=1)
+            if response.peek_stdout():
+                std_out.append(response.read_stdout())
+            if response.peek_stderr():
+                std_err.append(response.read_stderr())
+            if cmd:
+                c = cmd.pop(0)
+                response.write_stdin(c + "\n")
+            else:
+                break
+        if std_out:
+            return std_out
